#include "atarifilesystem.h"

QString AtariDirEntry::name()
{
    QString s = baseName();
    QString e = suffix();
    if (!e.isEmpty()) {
        s.append(".");
        s.append(e);
    }
    return s;
}

QString AtariDirEntry::baseName()
{
    return QString::fromAscii(atariName.left(8).constData()).trimmed();
}

QString AtariDirEntry::suffix()
{
    return QString::fromAscii(atariName.right(3).constData()).trimmed();
}


AtariFileSystem::AtariFileSystem(DiskImage *image)
{
    QByteArray dir, d;
    image->readSector(361, d); dir.append(d);
    image->readSector(362, d); dir.append(d);
    image->readSector(363, d); dir.append(d);
    image->readSector(364, d); dir.append(d);
    image->readSector(365, d); dir.append(d);
    image->readSector(366, d); dir.append(d);
    image->readSector(367, d); dir.append(d);
    image->readSector(368, d); dir.append(d);

    for (int i = 0; i < 1024; i += 16) {
        Dos20Entry e;
        e.internalData = dir.mid(i, 16);
        AtariDirEntry de = e.decoded();
        if (!(de.attributes & AtariDirEntry::InUse)) {
            continue;
        }
        //qDebug(de.name().toLower().toLocal8Bit());
    }
}

AtariDirEntry Dos20Entry::decoded()
{
    AtariDirEntry e;
    e.attributes = attributes();
    e.atariName = internalData.mid(5, 11);
    e.numberOfBlocks = numberOfBlocks();
    e.dateTime = QDateTime();
    return e;
}

AtariDirEntry::Attributes Dos20Entry::attributes()
{
    AtariDirEntry::Attributes a = 0;

    quint8 f = internalData.at(0);

    if (f & 0x01) {
        if (f & 0x40) {
            a |= AtariDirEntry::OpenForWrite;
        } else {
            a |= AtariDirEntry::Dos25;
            a |= AtariDirEntry::InUse;
        }
    }
    if (!(f & 0x02)) {
        a |= AtariDirEntry::Dos10;
    }
    if (f & 0x04) {
        a |= AtariDirEntry::MyDos;
    }
    if (f & 0x10) {
        a |= AtariDirEntry::Directory;
    }
    if (f & 0x20) {
        a |= AtariDirEntry::Locked;
    }
    if (f & 0x40) {
        a |= AtariDirEntry::InUse;
    }
    if (f & 0x80) {
        a |= AtariDirEntry::Deleted;
    }

    return a;
}

quint16 Dos20Entry::numberOfBlocks()
{
    return internalData.at(1) + internalData.at(2) * 256;
}

quint16 Dos20Entry::firstSector()
{
    return internalData.at(3) + internalData.at(4) * 256;
}
