#include "mainwindow.h"
#include "ui_mainwindow.h"

#include "diskimage.h"
#include "folderimage.h"
#include "miscdevices.h"
#include "aspeqtsettings.h"
#include "autobootdialog.h"
#include "autoboot.h"

#include <QDragEnterEvent>
#include <QDropEvent>
#include <QUrl>
#include <QScrollBar>

#include "atarifilesystem.h"

AspeqtSettings *aspeqtSettings;

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent), ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    speedLabel = new QLabel(this);
    onOffLabel = new QLabel(this);
    clearMessagesLabel = new QLabel(this);

    speedLabel->setText(tr("19200 kbps"));

    onOffLabel->setMinimumWidth(21);

    clearMessagesLabel->setMinimumWidth(21);
    clearMessagesLabel->setPixmap(QIcon(":/icons/silk-icons/icons/page_white.png").pixmap(16, 16, QIcon::Normal));
    clearMessagesLabel->setToolTip(tr("Clear messages"));
    clearMessagesLabel->setStatusTip(clearMessagesLabel->toolTip());

    speedLabel->setMinimumWidth(60);

    ui->statusBar->addPermanentWidget(speedLabel);
    ui->statusBar->addPermanentWidget(onOffLabel);
    ui->statusBar->addPermanentWidget(clearMessagesLabel);

    /* Set application properties */
    QCoreApplication::setOrganizationName("TrayGun Software");
    QCoreApplication::setApplicationName("AspeQt");
    aspeqtSettings = new AspeqtSettings();

    /* Set tool button actions */
    ui->buttonMountDisk_1->setDefaultAction(ui->actionMountDisk_1);
    ui->buttonMountDisk_2->setDefaultAction(ui->actionMountDisk_2);
    ui->buttonMountDisk_3->setDefaultAction(ui->actionMountDisk_3);
    ui->buttonMountDisk_4->setDefaultAction(ui->actionMountDisk_4);
    ui->buttonMountDisk_5->setDefaultAction(ui->actionMountDisk_5);
    ui->buttonMountDisk_6->setDefaultAction(ui->actionMountDisk_6);
    ui->buttonMountDisk_7->setDefaultAction(ui->actionMountDisk_7);
    ui->buttonMountDisk_8->setDefaultAction(ui->actionMountDisk_8);
    ui->buttonMountFolder_1->setDefaultAction(ui->actionMountFolder_1);
    ui->buttonMountFolder_2->setDefaultAction(ui->actionMountFolder_2);
    ui->buttonMountFolder_3->setDefaultAction(ui->actionMountFolder_3);
    ui->buttonMountFolder_4->setDefaultAction(ui->actionMountFolder_4);
    ui->buttonMountFolder_5->setDefaultAction(ui->actionMountFolder_5);
    ui->buttonMountFolder_6->setDefaultAction(ui->actionMountFolder_6);
    ui->buttonMountFolder_7->setDefaultAction(ui->actionMountFolder_7);
    ui->buttonMountFolder_8->setDefaultAction(ui->actionMountFolder_8);
    ui->buttonEject_1->setDefaultAction(ui->actionEject_1);
    ui->buttonEject_2->setDefaultAction(ui->actionEject_2);
    ui->buttonEject_3->setDefaultAction(ui->actionEject_3);
    ui->buttonEject_4->setDefaultAction(ui->actionEject_4);
    ui->buttonEject_5->setDefaultAction(ui->actionEject_5);
    ui->buttonEject_6->setDefaultAction(ui->actionEject_6);
    ui->buttonEject_7->setDefaultAction(ui->actionEject_7);
    ui->buttonEject_8->setDefaultAction(ui->actionEject_8);
    ui->buttonWriteProtect_1->setDefaultAction(ui->actionWriteProtect_1);
    ui->buttonWriteProtect_2->setDefaultAction(ui->actionWriteProtect_2);
    ui->buttonWriteProtect_3->setDefaultAction(ui->actionWriteProtect_3);
    ui->buttonWriteProtect_4->setDefaultAction(ui->actionWriteProtect_4);
    ui->buttonWriteProtect_5->setDefaultAction(ui->actionWriteProtect_5);
    ui->buttonWriteProtect_6->setDefaultAction(ui->actionWriteProtect_6);
    ui->buttonWriteProtect_7->setDefaultAction(ui->actionWriteProtect_7);
    ui->buttonWriteProtect_8->setDefaultAction(ui->actionWriteProtect_8);
    ui->buttonEditDisk_1->setDefaultAction(ui->actionEditDisk_1);
    ui->buttonEditDisk_2->setDefaultAction(ui->actionEditDisk_2);
    ui->buttonEditDisk_3->setDefaultAction(ui->actionEditDisk_3);
    ui->buttonEditDisk_4->setDefaultAction(ui->actionEditDisk_4);
    ui->buttonEditDisk_5->setDefaultAction(ui->actionEditDisk_5);
    ui->buttonEditDisk_6->setDefaultAction(ui->actionEditDisk_6);
    ui->buttonEditDisk_7->setDefaultAction(ui->actionEditDisk_7);
    ui->buttonEditDisk_8->setDefaultAction(ui->actionEditDisk_8);

    /* Initialize disk widget array */
    diskWidgets[0].fileNameLabel = ui->labelFileName_1;
    diskWidgets[1].fileNameLabel = ui->labelFileName_2;
    diskWidgets[2].fileNameLabel = ui->labelFileName_3;
    diskWidgets[3].fileNameLabel = ui->labelFileName_4;
    diskWidgets[4].fileNameLabel = ui->labelFileName_5;
    diskWidgets[5].fileNameLabel = ui->labelFileName_6;
    diskWidgets[6].fileNameLabel = ui->labelFileName_7;
    diskWidgets[7].fileNameLabel = ui->labelFileName_8;

    diskWidgets[0].imagePropertiesLabel = ui->labelImageProperties_1;
    diskWidgets[1].imagePropertiesLabel = ui->labelImageProperties_2;
    diskWidgets[2].imagePropertiesLabel = ui->labelImageProperties_3;
    diskWidgets[3].imagePropertiesLabel = ui->labelImageProperties_4;
    diskWidgets[4].imagePropertiesLabel = ui->labelImageProperties_5;
    diskWidgets[5].imagePropertiesLabel = ui->labelImageProperties_6;
    diskWidgets[6].imagePropertiesLabel = ui->labelImageProperties_7;
    diskWidgets[7].imagePropertiesLabel = ui->labelImageProperties_8;

    diskWidgets[0].ejectAction = ui->actionEject_1;
    diskWidgets[1].ejectAction = ui->actionEject_2;
    diskWidgets[2].ejectAction = ui->actionEject_3;
    diskWidgets[3].ejectAction = ui->actionEject_4;
    diskWidgets[4].ejectAction = ui->actionEject_5;
    diskWidgets[5].ejectAction = ui->actionEject_6;
    diskWidgets[6].ejectAction = ui->actionEject_7;
    diskWidgets[7].ejectAction = ui->actionEject_8;

    diskWidgets[0].writeProtectAction = ui->actionWriteProtect_1;
    diskWidgets[1].writeProtectAction = ui->actionWriteProtect_2;
    diskWidgets[2].writeProtectAction = ui->actionWriteProtect_3;
    diskWidgets[3].writeProtectAction = ui->actionWriteProtect_4;
    diskWidgets[4].writeProtectAction = ui->actionWriteProtect_5;
    diskWidgets[5].writeProtectAction = ui->actionWriteProtect_6;
    diskWidgets[6].writeProtectAction = ui->actionWriteProtect_7;
    diskWidgets[7].writeProtectAction = ui->actionWriteProtect_8;

    diskWidgets[0].editAction = ui->actionEditDisk_1;
    diskWidgets[1].editAction = ui->actionEditDisk_2;
    diskWidgets[2].editAction = ui->actionEditDisk_3;
    diskWidgets[3].editAction = ui->actionEditDisk_4;
    diskWidgets[4].editAction = ui->actionEditDisk_5;
    diskWidgets[5].editAction = ui->actionEditDisk_6;
    diskWidgets[6].editAction = ui->actionEditDisk_7;
    diskWidgets[7].editAction = ui->actionEditDisk_8;

    diskWidgets[0].frame = ui->horizontalFrame_1;
    diskWidgets[1].frame = ui->horizontalFrame_2;
    diskWidgets[2].frame = ui->horizontalFrame_3;
    diskWidgets[3].frame = ui->horizontalFrame_4;
    diskWidgets[4].frame = ui->horizontalFrame_5;
    diskWidgets[5].frame = ui->horizontalFrame_6;
    diskWidgets[6].frame = ui->horizontalFrame_7;
    diskWidgets[7].frame = ui->horizontalFrame_8;

    /* Connect SioWorker signals */
    qRegisterMetaType<MessageType::UiMessageType>("MessageType::UiMessageType");
    sio = new SioWorker();
    connect(sio, SIGNAL(started()), this, SLOT(sioStarted()));
    connect(sio, SIGNAL(finished()), this, SLOT(sioFinished()));
    connect(sio, SIGNAL(speedChanged(uint)), this, SLOT(sioSpeedChanged(uint)));
    connect(sio, SIGNAL(uiMessage(const QString, MessageType::UiMessageType)), this, SLOT(uiMessage(QString, MessageType::UiMessageType)));
    shownFirstTime = true;

    /* Restore application state */
    for (int i = 0; i < 8; i++) {
        diskWidgets[i].editAction->setVisible(false);
        AspeqtSettings::ImageSettings is;
        is = aspeqtSettings->mountedImageSetting(i);
        mountFile(i, is.fileName, is.isWriteProtected);
    }
    updateRecentFileActions();

    setAcceptDrops(true);

    ApeTime *ape = new ApeTime(sio);
    sio->installDevice(0x45, ape);
}

MainWindow::~MainWindow()
{
    if (ui->actionStartEmulation->isChecked()) {
        ui->actionStartEmulation->trigger();
    }

    delete aspeqtSettings;
    delete sio;

    delete ui;
}

 void MainWindow::mousePressEvent(QMouseEvent *event)
 {
     int slot = containingDiskSlot(event->pos());

     if (event->button() == Qt::LeftButton
         && slot >= 0) {

         QDrag *drag = new QDrag(this);
         QMimeData *mimeData = new QMimeData;

         mimeData->setData("application/x-aspeqt-disk-image", QByteArray(1, slot));
         drag->setMimeData(mimeData);

         drag->exec();
     }

     if (event->button() == Qt::LeftButton && onOffLabel->geometry().translated(ui->statusBar->geometry().topLeft()).contains(event->pos())) {
         ui->actionStartEmulation->trigger();
     }

      if (event->button() == Qt::LeftButton && clearMessagesLabel->geometry().translated(ui->statusBar->geometry().topLeft()).contains(event->pos())) {
         ui->textEdit->clear();
     }
}

void MainWindow::dragEnterEvent(QDragEnterEvent *event)
{
    int i = containingDiskSlot(event->pos());
    if (i >= 0 && (event->mimeData()->hasUrls() ||
                   event->mimeData()->hasFormat("application/x-aspeqt-disk-image"))) {
        event->acceptProposedAction();
    } else {
        i = -1;
    }
    for (int j = 0; j <= 7; j++) {
        if (i == j) {
            diskWidgets[j].frame->setFrameShadow(QFrame::Sunken);
        } else {
            diskWidgets[j].frame->setFrameShadow(QFrame::Raised);
        }
    }
}

void MainWindow::dragMoveEvent(QDragMoveEvent *event)
{
    int i = containingDiskSlot(event->pos());
    if (i >= 0 && (event->mimeData()->hasUrls() ||
                   event->mimeData()->hasFormat("application/x-aspeqt-disk-image"))) {
        event->acceptProposedAction();
    } else {
        i = -1;
    }
    for (int j = 0; j <= 7; j++) {
        if (i == j) {
            diskWidgets[j].frame->setFrameShadow(QFrame::Sunken);
        } else {
            diskWidgets[j].frame->setFrameShadow(QFrame::Raised);
        }
    }
}

void MainWindow::dragLeaveEvent(QDragLeaveEvent *)
{
    for (int j = 0; j <= 7; j++) {
        diskWidgets[j].frame->setFrameShadow(QFrame::Raised);
    }
}

void MainWindow::dropEvent(QDropEvent *event)
{
    for (int j = 0; j <= 7; j++) {
        diskWidgets[j].frame->setFrameShadow(QFrame::Raised);
    }
    int slot = containingDiskSlot(event->pos());
    if (!(event->mimeData()->hasUrls() ||
          event->mimeData()->hasFormat("application/x-aspeqt-disk-image")) ||
          slot < 0) {
        return;
    }

    if (event->mimeData()->hasFormat("application/x-aspeqt-disk-image")) {
        int source = event->mimeData()->data("application/x-aspeqt-disk-image").at(0);

        if (slot == source) {
            return;
        }

        sio->swapDevices(slot + 0x31, source + 0x31);

        QString tempStr = diskWidgets[slot].fileNameLabel->text();
        diskWidgets[slot].fileNameLabel->setText(diskWidgets[source].fileNameLabel->text());
        diskWidgets[source].fileNameLabel->setText(tempStr);

        bool tempBool = diskWidgets[slot].ejectAction->isEnabled();
        diskWidgets[slot].ejectAction->setEnabled(diskWidgets[source].ejectAction->isEnabled());
        diskWidgets[source].ejectAction->setEnabled(tempBool);

        tempBool = diskWidgets[slot].writeProtectAction->isEnabled();
        diskWidgets[slot].writeProtectAction->setEnabled(diskWidgets[source].writeProtectAction->isEnabled());
        diskWidgets[source].writeProtectAction->setEnabled(tempBool);

        tempBool = diskWidgets[slot].writeProtectAction->isChecked();
        diskWidgets[slot].writeProtectAction->setChecked(diskWidgets[source].writeProtectAction->isChecked());
        diskWidgets[source].writeProtectAction->setChecked(tempBool);

        aspeqtSettings->swapImages(slot, source);

        uiMessage(tr("Swapped disk %1 with disk %2.").arg(slot + 1).arg(source + 1), MessageType::NormalMessage);

        return;
    }

    QStringList files;
    foreach (QUrl url, event->mimeData()->urls()) {
        if (!url.toLocalFile().isEmpty()) {
            files.append(url.toLocalFile());
        }
    }
    if (files.isEmpty()) {
        return;
    }
    mountFileWithDefaultProtection(slot, files[0]);
    files.removeAt(0);
    while (!files.isEmpty() && (slot = firstEmptyDiskSlot(slot, false)) >= 0) {
        mountFileWithDefaultProtection(slot, files[0]);
        files.removeAt(0);
    }
    slot = 0;
    while (!files.isEmpty() && (slot = firstEmptyDiskSlot(slot, false)) >= 0) {
        mountFileWithDefaultProtection(slot, files[0]);
        files.removeAt(0);
    }
    foreach(QString file, files) {
        uiMessage(tr("Cannot mount '%1': No empty disk slots.").arg(file), MessageType::ErrorMessage);
    }
}

void MainWindow::show()
{
    QMainWindow::show();
    if (shownFirstTime) {
        /* Open options dialog if it's first the time */
        if (aspeqtSettings->isFirstTime()) {
            if (QMessageBox::Yes == QMessageBox::question(this, tr("First run"),
                                       tr("You are running AspeQt for the first time.\n\nDo you want to open the options dialog?"),
                                       QMessageBox::Yes, QMessageBox::No)) {
                ui->actionOptions->trigger();
            }
        }
        ui->actionStartEmulation->trigger();
    }
}

void MainWindow::on_actionStartEmulation_triggered()
{
    if (ui->actionStartEmulation->isChecked()) {
        sio->start(QThread::TimeCriticalPriority);
    } else {
        sio->setPriority(QThread::NormalPriority);
        sio->wait();
    }
}

void MainWindow::sioStarted()
{
    ui->actionStartEmulation->setText(tr("&Stop emulation"));
    ui->actionStartEmulation->setToolTip(tr("Stop SIO peripheral emulation"));
    ui->actionStartEmulation->setStatusTip(tr("Stop SIO peripheral emulation"));
    onOffLabel->setPixmap(ui->actionStartEmulation->icon().pixmap(16, QIcon::Normal, QIcon::On));
    onOffLabel->setToolTip(ui->actionStartEmulation->toolTip());
    onOffLabel->setStatusTip(ui->actionStartEmulation->statusTip());
}

void MainWindow::sioFinished()
{
    ui->actionStartEmulation->setText(tr("&Start emulation"));
    ui->actionStartEmulation->setToolTip(tr("Start SIO peripheral emulation"));
    ui->actionStartEmulation->setStatusTip(tr("Start SIO peripheral emulation"));
    onOffLabel->setPixmap(ui->actionStartEmulation->icon().pixmap(16, QIcon::Normal, QIcon::Off));
    onOffLabel->setToolTip(ui->actionStartEmulation->toolTip());
    onOffLabel->setStatusTip(ui->actionStartEmulation->statusTip());
    speedLabel->hide();
    speedLabel->clear();
    uiMessage(tr("Emulation stopped."), MessageType::ImportantMessage);
}

void MainWindow::sioSpeedChanged(uint newSpeed)
{
    speedLabel->setText(tr("%1 kbps").arg(newSpeed));
    speedLabel->show();
}

void MainWindow::deviceStatusChanged(int deviceNo)
{
    if (deviceNo >= 0x31 && deviceNo <= 0x38) {
        DiskImage *img = qobject_cast <DiskImage*> (sio->getDevice(deviceNo));
        if (img) {
            diskWidgets[deviceNo - 0x31].imagePropertiesLabel->setText(img->properties());
        } else {
            diskWidgets[deviceNo - 0x31].imagePropertiesLabel->clear();
        }
    }
}

void MainWindow::uiMessage(const QString message, MessageType::UiMessageType type)
{
    if (type < MessageType::UnimportantMessage) {
        return;
    }
    switch (type) {
        case MessageType::DebugMessage:
            return;
            break;
        case MessageType::UnimportantMessage:
            ui->textEdit->append(QString("<span style='color:gray'>%1</span>").arg(message));
            break;

        case MessageType::NormalMessage:
            ui->textEdit->append(QString("<span style='color:black'>%1</span>").arg(message));
            break;

        case MessageType::ImportantMessage:
            ui->textEdit->append(QString("<span style='color:blue'>%1</span>").arg(message));
            break;

        case MessageType::WarningMessage:
            ui->textEdit->append(QString("<span style='color:orange'>%1</span>").arg(message));
            break;

        default:
        case MessageType::ErrorMessage:
            ui->textEdit->append(QString("<span style='color:red'>%1</span>").arg(message));
            break;
    }

    ui->statusBar->showMessage(message, 3000);
    ui->textEdit->verticalScrollBar()->setSliderPosition(ui->textEdit->verticalScrollBar()->maximum());
    ui->textEdit->horizontalScrollBar()->setSliderPosition(ui->textEdit->horizontalScrollBar()->minimum());
}

void MainWindow::on_actionOptions_triggered()
{
    bool restart;
    restart = ui->actionStartEmulation->isChecked();
    if (restart) {
        ui->actionStartEmulation->trigger();
    }
    OptionsDialog optionsDialog(this);
    optionsDialog.exec();
    if (restart) {
        ui->actionStartEmulation->trigger();
    }
}

void MainWindow::updateRecentFileActions()
{
    ui->actionMountRecent_0->setText(aspeqtSettings->recentImageSetting(0).fileName);
    ui->actionMountRecent_1->setText(aspeqtSettings->recentImageSetting(1).fileName);
    ui->actionMountRecent_2->setText(aspeqtSettings->recentImageSetting(2).fileName);
    ui->actionMountRecent_3->setText(aspeqtSettings->recentImageSetting(3).fileName);
    ui->actionMountRecent_4->setText(aspeqtSettings->recentImageSetting(4).fileName);
    ui->actionMountRecent_5->setText(aspeqtSettings->recentImageSetting(5).fileName);
    ui->actionMountRecent_6->setText(aspeqtSettings->recentImageSetting(6).fileName);
    ui->actionMountRecent_7->setText(aspeqtSettings->recentImageSetting(7).fileName);
    ui->actionMountRecent_8->setText(aspeqtSettings->recentImageSetting(8).fileName);
    ui->actionMountRecent_9->setText(aspeqtSettings->recentImageSetting(9).fileName);

    ui->actionMountRecent_0->setVisible(!ui->actionMountRecent_0->text().isEmpty());
    ui->actionMountRecent_1->setVisible(!ui->actionMountRecent_1->text().isEmpty());
    ui->actionMountRecent_2->setVisible(!ui->actionMountRecent_2->text().isEmpty());
    ui->actionMountRecent_3->setVisible(!ui->actionMountRecent_3->text().isEmpty());
    ui->actionMountRecent_4->setVisible(!ui->actionMountRecent_4->text().isEmpty());
    ui->actionMountRecent_5->setVisible(!ui->actionMountRecent_5->text().isEmpty());
    ui->actionMountRecent_6->setVisible(!ui->actionMountRecent_6->text().isEmpty());
    ui->actionMountRecent_7->setVisible(!ui->actionMountRecent_7->text().isEmpty());
    ui->actionMountRecent_8->setVisible(!ui->actionMountRecent_8->text().isEmpty());
    ui->actionMountRecent_9->setVisible(!ui->actionMountRecent_9->text().isEmpty());
}


void MainWindow::ejectImage(int no)
{
    SioDevice *dev = sio->getDevice(no + 0x31);
    sio->uninstallDevice(no + 0x31);
    if (!dev) {
        return;
    }
    delete dev;
    diskWidgets[no].ejectAction->setEnabled(false);
    QString fileName = diskWidgets[no].fileNameLabel->text();
    diskWidgets[no].fileNameLabel->clear();
    diskWidgets[no].writeProtectAction->setChecked(false);
    diskWidgets[no].writeProtectAction->setEnabled(false);
    diskWidgets[no].editAction->setEnabled(false);

    aspeqtSettings->unmountImage(no);
    updateRecentFileActions();
    deviceStatusChanged(no + 0x31);
    uiMessage(tr("Ejected disk %1").arg(no + 1), MessageType::NormalMessage);
}

int MainWindow::containingDiskSlot(const QPoint &point)
{
    int i;
    QPoint distance = centralWidget()->geometry().topLeft();
    for (i=0; i < 8; i++) {
        QRect rect = diskWidgets[i].frame->geometry().translated(distance);
        if (rect.contains(point)) {
            break;
        }
    }
    if (i > 7) {
        i = -1;
    }
    return i;
}

int MainWindow::firstEmptyDiskSlot(int startFrom, bool createOne)
{
    int i;
    for (i = startFrom; i < 8; i++) {
        if (!sio->getDevice(0x31 + i)) {
            break;
        }
    }
    if (i > 7) {
        if (createOne) {
            i = 7;
        } else {
            i = -1;
        }
    }
    return i;
}

void MainWindow::mountFileWithDefaultProtection(int no, const QString &fileName)
{
    bool prot = aspeqtSettings->getImageSettingsFromName(fileName).isWriteProtected;
    mountFile(no, fileName, prot);
}

void MainWindow::mountFile(int no, const QString &fileName, bool prot)
{
    DiskImage *disk;
    DiskImage::CompressionType compression = DiskImage::NoCompression;

    if (fileName.isEmpty()) {
        return;
    }
    else if (QFileInfo(fileName).isDir()) {
        disk = new FolderImage(sio);
    } else if (fileName.endsWith(".ATR", Qt::CaseInsensitive)) {
        disk = new AtrImage(sio);
    } else if (fileName.endsWith(".XFD", Qt::CaseInsensitive)) {
        disk = new XfdImage(sio);
    } else if (fileName.endsWith(".DCM", Qt::CaseInsensitive)) {
        uiMessage(tr("Cannot mount '%1': Mounting DCM images is not implemented yet.").arg(fileName), MessageType::ErrorMessage);
        disk = 0;
    } else if (fileName.endsWith(".SCP", Qt::CaseInsensitive)) {
        uiMessage(tr("Cannot mount '%1': Mounting SCP images is not implemented yet.").arg(fileName), MessageType::ErrorMessage);
        disk = 0;
    } else if (fileName.endsWith(".ATZ", Qt::CaseInsensitive) || fileName.endsWith(".ATR.GZ", Qt::CaseInsensitive)) {
        disk = new AtrImage(sio);
        compression = DiskImage::GzipCompression;
    } else if (fileName.endsWith(".XFZ", Qt::CaseInsensitive) || fileName.endsWith(".XFD.GZ", Qt::CaseInsensitive)) {
        disk = new XfdImage(sio);
        compression = DiskImage::GzipCompression;
    } else {
        disk = new RawDiskImage(sio);
    }

    if (disk) {
        if (!disk->open(fileName, compression, false)) {
            delete disk;
            return;
        }
        ejectImage(no);

        sio->installDevice(0x31 + no, disk);
        diskWidgets[no].ejectAction->setEnabled(true);
        diskWidgets[no].editAction->setEnabled(true);
        diskWidgets[no].writeProtectAction->setEnabled(!disk->isReadOnly());
        diskWidgets[no].writeProtectAction->setChecked(disk->isReadOnly());
        if (!disk->isReadOnly() && prot) {
            diskWidgets[no].writeProtectAction->trigger();
        }
        diskWidgets[no].fileNameLabel->setText(fileName);
        aspeqtSettings->mountImage(no, fileName, disk->isReadOnly());
        updateRecentFileActions();
        connect(disk, SIGNAL(statusChanged(int)), this, SLOT(deviceStatusChanged(int)), Qt::QueuedConnection);
        deviceStatusChanged(0x31 + no);
        uiMessage(tr("[%1] Mounted '%2' as %3.")
                  .arg(disk->deviceName())
                  .arg(fileName)
                  .arg(disk->properties()), MessageType::NormalMessage);
    }
}

void MainWindow::mountDiskImage(int no)
{
    QString dir;
    if (diskWidgets[no].fileNameLabel->text().isEmpty()) {
        dir = aspeqtSettings->lastDiskImageDir();
    } else {
        dir = QFileInfo(diskWidgets[no].fileNameLabel->text()).absolutePath();
    }
    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Open a disk image"),
                                                    dir,
                                                    tr(
                                                    "All Atari disk images (*.atr *.xfd *.dcm *.scp *.atz *.xfz *.atr.gz *.xfd.gz);;"
                                                    "SIO2PC ATR images (*.atr);;"
                                                    "Compressed ATR images (*.atz *.atr.gz);;"
                                                    "SpartaDos SCOPY SCP images (*.scp);;"
                                                    "XFormer XFD images (*.xfd);;"
                                                    "Compressed XFD images (*.xfz *.xfd.gz);;"
                                                    "DiskComm DCM images (*.dcm);;"
                                                    "All files (*)"));
    if (fileName.isEmpty()) {
        return;
    }
    aspeqtSettings->setLastDiskImageDir(QFileInfo(fileName).absolutePath());
    mountFileWithDefaultProtection(no, fileName);
}

void MainWindow::mountFolderImage(int no)
{
    QString dir;
    if (diskWidgets[no].fileNameLabel->text().isEmpty()) {
        dir = aspeqtSettings->lastFolderImageDir();
    } else {
        dir = diskWidgets[no].fileNameLabel->text();
    }
    QString fileName = QFileDialog::getExistingDirectory(this, tr("Open a folder image"), dir);
    if (fileName.isEmpty()) {
        return;
    }
    aspeqtSettings->setLastFolderImageDir(fileName);
    mountFileWithDefaultProtection(no, fileName);
}

void MainWindow::toggleWriteProtection(int no)
{
    DiskImage *img = qobject_cast <DiskImage*> (sio->getDevice(no + 0x31));
    if (diskWidgets[no].writeProtectAction->isChecked()) {
        img->setReadOnly(true);
    } else {
        img->setReadOnly(false);
    }
    aspeqtSettings->setMountedImageSetting(no, diskWidgets[no].fileNameLabel->text(), diskWidgets[no].writeProtectAction->isChecked());
}

void MainWindow::openEditor(int no)
{
}

void MainWindow::on_actionMountDisk_1_triggered() {mountDiskImage(0);}
void MainWindow::on_actionMountDisk_2_triggered() {mountDiskImage(1);}
void MainWindow::on_actionMountDisk_3_triggered() {mountDiskImage(2);}
void MainWindow::on_actionMountDisk_4_triggered() {mountDiskImage(3);}
void MainWindow::on_actionMountDisk_5_triggered() {mountDiskImage(4);}
void MainWindow::on_actionMountDisk_6_triggered() {mountDiskImage(5);}
void MainWindow::on_actionMountDisk_7_triggered() {mountDiskImage(6);}
void MainWindow::on_actionMountDisk_8_triggered() {mountDiskImage(7);}

void MainWindow::on_actionMountFolder_1_triggered() {mountFolderImage(0);}
void MainWindow::on_actionMountFolder_2_triggered() {mountFolderImage(1);}
void MainWindow::on_actionMountFolder_3_triggered() {mountFolderImage(2);}
void MainWindow::on_actionMountFolder_4_triggered() {mountFolderImage(3);}
void MainWindow::on_actionMountFolder_5_triggered() {mountFolderImage(4);}
void MainWindow::on_actionMountFolder_6_triggered() {mountFolderImage(5);}
void MainWindow::on_actionMountFolder_7_triggered() {mountFolderImage(6);}
void MainWindow::on_actionMountFolder_8_triggered() {mountFolderImage(7);}

void MainWindow::on_actionEject_1_triggered() {ejectImage(0);}
void MainWindow::on_actionEject_2_triggered() {ejectImage(1);}
void MainWindow::on_actionEject_3_triggered() {ejectImage(2);}
void MainWindow::on_actionEject_4_triggered() {ejectImage(3);}
void MainWindow::on_actionEject_5_triggered() {ejectImage(4);}
void MainWindow::on_actionEject_6_triggered() {ejectImage(5);}
void MainWindow::on_actionEject_7_triggered() {ejectImage(6);}
void MainWindow::on_actionEject_8_triggered() {ejectImage(7);}

void MainWindow::on_actionWriteProtect_1_triggered() {toggleWriteProtection(0);}
void MainWindow::on_actionWriteProtect_2_triggered() {toggleWriteProtection(1);}
void MainWindow::on_actionWriteProtect_3_triggered() {toggleWriteProtection(2);}
void MainWindow::on_actionWriteProtect_4_triggered() {toggleWriteProtection(3);}
void MainWindow::on_actionWriteProtect_5_triggered() {toggleWriteProtection(4);}
void MainWindow::on_actionWriteProtect_6_triggered() {toggleWriteProtection(5);}
void MainWindow::on_actionWriteProtect_7_triggered() {toggleWriteProtection(6);}
void MainWindow::on_actionWriteProtect_8_triggered() {toggleWriteProtection(7);}

void MainWindow::on_actionMountRecent_0_triggered() {mountFileWithDefaultProtection(firstEmptyDiskSlot(), ui->actionMountRecent_0->text());}
void MainWindow::on_actionMountRecent_1_triggered() {mountFileWithDefaultProtection(firstEmptyDiskSlot(), ui->actionMountRecent_1->text());}
void MainWindow::on_actionMountRecent_2_triggered() {mountFileWithDefaultProtection(firstEmptyDiskSlot(), ui->actionMountRecent_2->text());}
void MainWindow::on_actionMountRecent_3_triggered() {mountFileWithDefaultProtection(firstEmptyDiskSlot(), ui->actionMountRecent_3->text());}
void MainWindow::on_actionMountRecent_4_triggered() {mountFileWithDefaultProtection(firstEmptyDiskSlot(), ui->actionMountRecent_4->text());}
void MainWindow::on_actionMountRecent_5_triggered() {mountFileWithDefaultProtection(firstEmptyDiskSlot(), ui->actionMountRecent_5->text());}
void MainWindow::on_actionMountRecent_6_triggered() {mountFileWithDefaultProtection(firstEmptyDiskSlot(), ui->actionMountRecent_6->text());}
void MainWindow::on_actionMountRecent_7_triggered() {mountFileWithDefaultProtection(firstEmptyDiskSlot(), ui->actionMountRecent_7->text());}
void MainWindow::on_actionMountRecent_8_triggered() {mountFileWithDefaultProtection(firstEmptyDiskSlot(), ui->actionMountRecent_8->text());}
void MainWindow::on_actionMountRecent_9_triggered() {mountFileWithDefaultProtection(firstEmptyDiskSlot(), ui->actionMountRecent_9->text());}

void MainWindow::on_actionEditDisk_1_triggered() {openEditor(0);}
void MainWindow::on_actionEditDisk_2_triggered() {openEditor(0);}
void MainWindow::on_actionEditDisk_3_triggered() {openEditor(0);}
void MainWindow::on_actionEditDisk_4_triggered() {openEditor(0);}
void MainWindow::on_actionEditDisk_5_triggered() {openEditor(0);}
void MainWindow::on_actionEditDisk_6_triggered() {openEditor(0);}
void MainWindow::on_actionEditDisk_7_triggered() {openEditor(0);}
void MainWindow::on_actionEditDisk_8_triggered() {openEditor(0);}

void MainWindow::on_actionEjectAll_triggered()
{
    for (int i=7; i>=0; i--) {
        ejectImage(i);
    }
}

void MainWindow::on_actionMountDisk_triggered()
{
    mountDiskImage(firstEmptyDiskSlot(0, true));
}

void MainWindow::on_actionMountFolder_triggered()
{
    mountFolderImage(firstEmptyDiskSlot(0, true));
}

void MainWindow::on_actionNewImage_triggered()
{
    CreateImageDialog dlg(this);
    if (!dlg.exec()) {
        return;
    };
    QString dir = aspeqtSettings->lastDiskImageDir();
    QString fileName = QFileDialog::getSaveFileName(this, tr("Save new image as"),
                                 dir,
                                 tr(
                                         "All Atari disk images (*.atr *.xfd *.dcm *.scp *.atz *.xfz *.atr.gz *.xfd.gz);;"
                                         "SIO2PC ATR images (*.atr);;"
                                         "Compressed ATR images (*.atz *.atr.gz);;"
                                         "SpartaDos SCOPY SCP images (*.scp);;"
                                         "XFormer XFD images (*.xfd);;"
                                         "Compressed XFD images (*.xfz *.xfd.gz);;"
                                         "DiskComm DCM images (*.dcm);;"
                                         "All files (*)"));
    if (fileName.isEmpty()) {
        return;
    }

    aspeqtSettings->setLastDiskImageDir(QFileInfo(fileName).absolutePath());

    RawDiskImage *disk;
    DiskImage::CompressionType compression;

    if (fileName.endsWith(".ATR", Qt::CaseInsensitive)) {
        disk = new AtrImage(sio);
        compression = DiskImage::NoCompression;
    } else if (fileName.endsWith(".XFD", Qt::CaseInsensitive)) {
        disk = new XfdImage(sio);
        compression = DiskImage::NoCompression;
    } else if (fileName.endsWith(".DCM", Qt::CaseInsensitive)) {
        uiMessage(tr("Cannot create '%1': DCM image support is not implemented yet.").arg(fileName), MessageType::ErrorMessage);
        return;
    } else if (fileName.endsWith(".SCP", Qt::CaseInsensitive)) {
        uiMessage(tr("Cannot create '%1': SCP image support is not implemented yet.").arg(fileName), MessageType::ErrorMessage);
        return;
    } else if (fileName.endsWith(".ATZ", Qt::CaseInsensitive) || fileName.endsWith(".ATR.GZ", Qt::CaseInsensitive)) {
        disk = new AtrImage(sio);
        compression = DiskImage::GzipCompression;
    } else if (fileName.endsWith(".XFZ", Qt::CaseInsensitive) || fileName.endsWith(".XFD.GZ", Qt::CaseInsensitive)) {
        disk = new XfdImage(sio);
        compression = DiskImage::GzipCompression;
    } else {
        disk = new RawDiskImage(sio);
        compression = DiskImage::NoCompression;
    }

    if (disk->create(fileName, dlg.sectorCount(), dlg.sectorSize(), compression)) {
        delete disk;
        mountFileWithDefaultProtection(firstEmptyDiskSlot(0, true), fileName);
    }
}

void MainWindow::on_actionOpenSession_triggered()
{
    QString dir = aspeqtSettings->lastSessionDir();
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open session"),
                                 dir,
                                 tr(
                                         "AspeQt sessions (*.aspeqt);;"
                                         "All files (*)"));
    if (fileName.isEmpty()) {
        return;
    }
    aspeqtSettings->setLastSessionDir(QFileInfo(fileName).absolutePath());
    aspeqtSettings->loadSessionFromFile(fileName);
    for (int i = 0; i < 8; i++) {
        AspeqtSettings::ImageSettings is;
        is = aspeqtSettings->mountedImageSetting(i);
        mountFile(i, is.fileName, is.isWriteProtected);
    }
}

void MainWindow::on_actionSaveSession_triggered()
{
    QString dir = aspeqtSettings->lastSessionDir();
    QString fileName = QFileDialog::getSaveFileName(this, tr("Save session as"),
                                 dir,
                                 tr(
                                         "AspeQt sessions (*.aspeqt);;"
                                         "All files (*)"));
    if (fileName.isEmpty()) {
        return;
    }
    aspeqtSettings->setLastSessionDir(QFileInfo(fileName).absolutePath());
    aspeqtSettings->saveSessionToFile(fileName);
}

void MainWindow::on_actionBootExe_triggered()
{
    QString dir = aspeqtSettings->lastExeDir();
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open executable"),
                                 dir,
                                 tr(
                                         "Atari executables (*.xex *.com *.exe);;"
                                         "All files (*)"));
    if (fileName.isEmpty()) {
        return;
    }
    aspeqtSettings->setLastExeDir(QFileInfo(fileName).absolutePath());

    SioDevice *old = sio->getDevice(0x31);
    AutoBoot loader(sio);
    AutoBootDialog dlg(this);

    if (!loader.open(fileName)) {
        return;
    }

    sio->uninstallDevice(0x31);
    sio->installDevice(0x31, &loader);
    connect(&loader, SIGNAL(sectorRead(int,int)), &dlg, SLOT(sectorRead(int,int)));
    dlg.exec();
    sio->uninstallDevice(0x31);
    if (old) {
        sio->installDevice(0x31, old);
        DiskImage *d = qobject_cast <DiskImage*> (old);
        d = qobject_cast <DiskImage*> (sio->getDevice(0x31));
    }
}
